/*******************************************************************************
 * Copyright (c) 2009 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.e4.languages.internal.javascript.debug.launching;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.debug.core.sourcelookup.AbstractSourceLookupParticipant;
import org.eclipse.e4.languages.javascript.debug.connect.JSONConstants;
import org.eclipse.e4.languages.javascript.debug.model.JSDIStackFrame;
import org.eclipse.e4.languages.javascript.jsdi.ScriptReference;
import org.eclipse.e4.languages.javascript.jsdi.StackFrame;
import org.osgi.framework.Constants;

/**
 * Default source lookup participant
 * 
 * @since 1.0
 */
public class JavascriptSourceLookupParticipant extends AbstractSourceLookupParticipant {

	/**
	 * 
	 */
	static final String OSGI_SYMBOLIC_NAME = "OSGi." + Constants.BUNDLE_SYMBOLICNAME; //$NON-NLS-1$
	static final Object[] NO_SOURCE = new Object[0];
	static final Object[] FILE = new Object[1];

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.sourcelookup.ISourceLookupParticipant#getSourceName(java.lang.Object)
	 */
	public String getSourceName(Object object) throws CoreException {
		if (object instanceof JSDIStackFrame) {
			return ((JSDIStackFrame) object).getUnderlyingStackFrame().location().scriptReference().sourcePath();
		}
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.sourcelookup.AbstractSourceLookupParticipant#findSourceElements(java.lang.Object)
	 */
	public Object[] findSourceElements(Object object) throws CoreException {
		if (object instanceof JSDIStackFrame) {
			JSDIStackFrame jsdiStackFrame = (JSDIStackFrame) object;
			StackFrame underlyingStackFrame = jsdiStackFrame.getUnderlyingStackFrame();
			ScriptReference script = underlyingStackFrame.location().scriptReference();
			if (script != null) {
				Map props = script.sourceProperties();
				IProject project = getProject(props);
				if (project != null) {
					String pathname = (String) props.get(JSONConstants.PATH);
					String name = (String) props.get(JSONConstants.NAME);
					if (pathname != null && name != null) {
						IPath path = new Path(pathname).append(name);
						if (path.getDevice() != null) {
							path = path.setDevice(null);
						}
						IFile file = project.getFile(path);
						if (file != null) {
							return new IFile[] { file };
						}
					}
				}
				return showExternalSource(script);
			}
		}
		return NO_SOURCE;
	}

	/**
	 * Tries to find the project based on the given set of source properties
	 * 
	 * @param properties
	 * @return the project to lookup in or <code>null</code> if no project can be found
	 */
	IProject getProject(Map properties) {
		IProject project = null;
		String bundle = (String) properties.get(OSGI_SYMBOLIC_NAME);
		if (bundle != null) {
			project = ResourcesPlugin.getWorkspace().getRoot().getProject(bundle);
			if (!project.isAccessible()) {
				bundle = (String) properties.get(Constants.BUNDLE_SYMBOLICNAME);
				if (bundle != null) {
					project = ResourcesPlugin.getWorkspace().getRoot().getProject(bundle);
					if (!project.isAccessible()) {
						return null;
					}
				} else {
					return null;
				}
			}
		} else {
			bundle = (String) properties.get(Constants.BUNDLE_SYMBOLICNAME);
			if (bundle != null) {
				project = ResourcesPlugin.getWorkspace().getRoot().getProject(bundle);
				if (!project.isAccessible()) {
					return null;
				}
			} else {
				return null;
			}
		}
		return project;
	}

	/**
	 * Shows the source in an external editor
	 * 
	 * @param script
	 * @return the collection of files to show in external editors
	 */
	Object[] showExternalSource(ScriptReference script) {
		try {
			String source = script.source();
			File tempdir = new File(System.getProperty("java.io.tmpdir")); //$NON-NLS-1$
			File file = new File(tempdir, script.sourcePath());
			file.deleteOnExit();
			FileWriter writer = new FileWriter(file);
			writer.write(source);
			writer.close();
			FILE[0] = file;
			return FILE;

		} catch (IOException e) {
			e.printStackTrace();
			return NO_SOURCE;
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.sourcelookup.AbstractSourceLookupParticipant#isFindDuplicates()
	 */
	public boolean isFindDuplicates() {
		return true;
	}
}
