/*******************************************************************************
 * Copyright (c) 2009 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/
package org.eclipse.e4.languages.javascript.debug.model;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.model.IValue;
import org.eclipse.debug.core.model.IVariable;
import org.eclipse.e4.languages.javascript.debug.connect.JSONConstants;
import org.eclipse.e4.languages.javascript.jsdi.ArrayReference;
import org.eclipse.e4.languages.javascript.jsdi.FunctionReference;
import org.eclipse.e4.languages.javascript.jsdi.NumberValue;
import org.eclipse.e4.languages.javascript.jsdi.PropertiesReference;
import org.eclipse.e4.languages.javascript.jsdi.Property;
import org.eclipse.e4.languages.javascript.jsdi.Value;

/**
 * Default implementation of {@link IValue}
 * 
 * @since 1.0
 */
public class JSDIValue extends JSDIDebugElement implements IValue {

	private Value value;
	private String name;
	private List properties;

	/**
	 * Constructor
	 * 
	 * @param variable
	 * @param value
	 */
	public JSDIValue(JSDIVariable variable, Value value) {
		super(variable.getJSDITarget());
		this.name = variable.getName();
		this.value = value;
	}

	/**
	 * Constructor
	 * 
	 * @param variable
	 * @param value
	 */
	public JSDIValue(JSDIProperty variable, Value value) {
		super(variable.getJSDITarget());
		this.name = variable.getName();
		this.value = value;
	}

	/**
	 * Hook to return the underlying value for this {@link JSDIValue}
	 * 
	 * @return the underlying {@link Value}
	 */
	protected Value getUnderlyingValue() {
		return this.value;
	}

	/**
	 * @return the detail string that is typically shown in the details pane
	 */
	public String getDetailString() {
		if (this.value instanceof ArrayReference) {
			ArrayReference array = (ArrayReference) this.value;
			return array.getValues().toString();
		}
		if (this.value instanceof FunctionReference) {
			// TODO this would be slick to show the function body code
			return ((FunctionReference) this.value).functionName();
		}
		if (this.value instanceof NumberValue) {
			NumberValue nvalue = (NumberValue) this.value;
			if (!nvalue.isNaN()) {
				return JSDIDebugModel.numberToString(nvalue.value());
			}
		}
		return this.value.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.IValue#getReferenceTypeName()
	 */
	public String getReferenceTypeName() throws DebugException {
		return this.value.getValueTypeName();
	}

	/**
	 * Return the name to display for the variable
	 * 
	 * @return the variable name
	 */
	public String getName() {
		return this.name;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.IValue#getValueString()
	 */
	public String getValueString() throws DebugException {
		StringBuffer buffer = new StringBuffer();
		if (this.value instanceof PropertiesReference) {
			// TODO resolve id - needs work for lifecycle
			buffer.append(this.value.toString()).append(JSONConstants.SPACE).append("(id=unknown)"); //$NON-NLS-1$
			return buffer.toString();
		}
		if (this.value instanceof NumberValue) {
			NumberValue nvalue = (NumberValue) this.value;
			if (!nvalue.isNaN()) {
				return JSDIDebugModel.numberToString(nvalue.value());
			}
		}
		return this.value.toString();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.IValue#getVariables()
	 */
	public synchronized IVariable[] getVariables() throws DebugException {
		if (!hasVariables()) {
			return null;
		}
		if (this.properties == null) {
			PropertiesReference reference = (PropertiesReference) this.value;
			List underlyingProperties = reference.properties();
			this.properties = new ArrayList(underlyingProperties.size());
			for (Iterator iterator = underlyingProperties.iterator(); iterator.hasNext();) {
				Property property = (Property) iterator.next();
				JSDIProperty jsdiProperty = new JSDIProperty(this, reference, property);
				this.properties.add(jsdiProperty);
			}
		}
		return (IVariable[]) this.properties.toArray(new IVariable[this.properties.size()]);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.IValue#hasVariables()
	 */
	public boolean hasVariables() throws DebugException {
		return (this.value instanceof PropertiesReference);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.debug.core.model.IValue#isAllocated()
	 */
	public boolean isAllocated() throws DebugException {
		return false;
	}

}
